"""                    GNU GENERAL PUBLIC LICENSE
                       Version 3, 29 June 2007

 Copyright (C) 2007 Free Software Foundation, Inc. <https://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.

                            Preamble

  The GNU General Public License is a free, copyleft license for
software and other kinds of works.

  The licenses for most software and other practical works are designed
to take away your freedom to share and change the works.  By contrast,
the GNU General Public License is intended to guarantee your freedom to
share and change all versions of a program--to make sure it remains free
software for all its users.  We, the Free Software Foundation, use the
GNU General Public License for most of our software; it applies also to
any other work released this way by its authors.  You can apply it to
your programs, too.

  When we speak of free software, we are referring to freedom, not
price.  Our General Public Licenses are designed to make sure that you
have the freedom to distribute copies of free software (and charge for
them if you wish), that you receive source code or can get it if you
want it, that you can change the software or use pieces of it in new
free programs, and that you know you can do these things.

  To protect your rights, we need to prevent others from denying you
these rights or asking you to surrender the rights.  Therefore, you have
certain responsibilities if you distribute copies of the software, or if
you modify it: responsibilities to respect the freedom of others.

  For example, if you distribute copies of such a program, whether
gratis or for a fee, you must pass on to the recipients the same
freedoms that you received.  You must make sure that they, too, receive
or can get the source code.  And you must show them these terms so they
know their rights.

  Developers that use the GNU GPL protect your rights with two steps:
(1) assert copyright on the software, and (2) offer you this License
giving you legal permission to copy, distribute and/or modify it.

  For the developers' and authors' protection, the GPL clearly explains
that there is no warranty for this free software.  For both users' and
authors' sake, the GPL requires that modified versions be marked as
changed, so that their problems will not be attributed erroneously to
authors of previous versions.

  Some devices are designed to deny users access to install or run
modified versions of the software inside them, although the manufacturer
can do so.  This is fundamentally incompatible with the aim of
protecting users' freedom to change the software.  The systematic
pattern of such abuse occurs in the area of products for individuals to
use, which is precisely where it is most unacceptable.  Therefore, we
have designed this version of the GPL to prohibit the practice for those
products.  If such problems arise substantially in other domains, we
stand ready to extend this provision to those domains in future versions
of the GPL, as needed to protect the freedom of users.

  Finally, every program is threatened constantly by software patents.
States should not allow patents to restrict development and use of
software on general-purpose computers, but in those that do, we wish to
avoid the special danger that patents applied to a free program could
make it effectively proprietary.  To prevent this, the GPL assures that
patents cannot be used to render the program non-free.

  The precise terms and conditions for copying, distribution and
modification follow.

                       TERMS AND CONDITIONS

  0. Definitions.

  "This License" refers to version 3 of the GNU General Public License.

  "Copyright" also means copyright-like laws that apply to other kinds of
works, such as semiconductor masks.

  "The Program" refers to any copyrightable work licensed under this
License.  Each licensee is addressed as "you".  "Licensees" and
"recipients" may be individuals or organizations.

  To "modify" a work means to copy from or adapt all or part of the work
in a fashion requiring copyright permission, other than the making of an
exact copy.  The resulting work is called a "modified version" of the
earlier work or a work "based on" the earlier work.

  A "covered work" means either the unmodified Program or a work based
on the Program.

  To "propagate" a work means to do anything with it that, without
permission, would make you directly or secondarily liable for
infringement under applicable copyright law, except executing it on a
computer or modifying a private copy.  Propagation includes copying,
distribution (with or without modification), making available to the
public, and in some countries other activities as well.

  To "convey" a work means any kind of propagation that enables other
parties to make or receive copies.  Mere interaction with a user through
a computer network, with no transfer of a copy, is not conveying.

  An interactive user interface displays "Appropriate Legal Notices"
to the extent that it includes a convenient and prominently visible
feature that (1) displays an appropriate copyright notice, and (2)
tells the user that there is no warranty for the work (except to the
extent that warranties are provided), that licensees may convey the
work under this License, and how to view a copy of this License.  If
the interface presents a list of user commands or options, such as a
menu, a prominent item in the list meets this criterion.

  1. Source Code.

  The "source code" for a work means the preferred form of the work
for making modifications to it.  "Object code" means any non-source
form of a work.

  A "Standard Interface" means an interface that either is an official
standard defined by a recognized standards body, or, in the case of
interfaces specified for a particular programming language, one that
is widely used among developers working in that language.

  The "System Libraries" of an executable work include anything, other
than the work as a whole, that (a) is included in the normal form of
packaging a Major Component, but which is not part of that Major
Component, and (b) serves only to enable use of the work with that
Major Component, or to implement a Standard Interface for which an
implementation is available to the public in source code form.  A
"Major Component", in this context, means a major essential component
(kernel, window system, and so on) of the specific operating system
(if any) on which the executable work runs, or a compiler used to
produce the work, or an object code interpreter used to run it.

  The "Corresponding Source" for a work in object code form means all
the source code needed to generate, install, and (for an executable
work) run the object code and to modify the work, including scripts to
control those activities.  However, it does not include the work's
System Libraries, or general-purpose tools or generally available free
programs which are used unmodified in performing those activities but
which are not part of the work.  For example, Corresponding Source
includes interface definition files associated with source files for
the work, and the source code for shared libraries and dynamically
linked subprograms that the work is specifically designed to require,
such as by intimate data communication or control flow between those
subprograms and other parts of the work.

  The Corresponding Source need not include anything that users
can regenerate automatically from other parts of the Corresponding
Source.

  The Corresponding Source for a work in source code form is that
same work.

  2. Basic Permissions.

  All rights granted under this License are granted for the term of
copyright on the Program, and are irrevocable provided the stated
conditions are met.  This License explicitly affirms your unlimited
permission to run the unmodified Program.  The output from running a
covered work is covered by this License only if the output, given its
content, constitutes a covered work.  This License acknowledges your
rights of fair use or other equivalent, as provided by copyright law.

  You may make, run and propagate covered works that you do not
convey, without conditions so long as your license otherwise remains
in force.  You may convey covered works to others for the sole purpose
of having them make modifications exclusively for you, or provide you
with facilities for running those works, provided that you comply with
the terms of this License in conveying all material for which you do
not control copyright.  Those thus making or running the covered works
for you must do so exclusively on your behalf, under your direction
and control, on terms that prohibit them from making any copies of
your copyrighted material outside their relationship with you.

  Conveying under any other circumstances is permitted solely under
the conditions stated below.  Sublicensing is not allowed; section 10
makes it unnecessary.

  3. Protecting Users' Legal Rights From Anti-Circumvention Law.

  No covered work shall be deemed part of an effective technological
measure under any applicable law fulfilling obligations under article
11 of the WIPO copyright treaty adopted on 20 December 1996, or
similar laws prohibiting or restricting circumvention of such
measures.

  When you convey a covered work, you waive any legal power to forbid
circumvention of technological measures to the extent such circumvention
is effected by exercising rights under this License with respect to
the covered work, and you disclaim any intention to limit operation or
modification of the work as a means of enforcing, against the work's
users, your or third parties' legal rights to forbid circumvention of
technological measures.

  4. Conveying Verbatim Copies.

  You may convey verbatim copies of the Program's source code as you
receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy an appropriate copyright notice;
keep intact all notices stating that this License and any
non-permissive terms added in accord with section 7 apply to the code;
keep intact all notices of the absence of any warranty; and give all
recipients a copy of this License along with the Program.

  You may charge any price or no price for each copy that you convey,
and you may offer support or warranty protection for a fee.

  5. Conveying Modified Source Versions.

  You may convey a work based on the Program, or the modifications to
produce it from the Program, in the form of source code under the
terms of section 4, provided that you also meet all of these conditions:

    a) The work must carry prominent notices stating that you modified
    it, and giving a relevant date.

    b) The work must carry prominent notices stating that it is
    released under this License and any conditions added under section
    7.  This requirement modifies the requirement in section 4 to
    "keep intact all notices".

    c) You must license the entire work, as a whole, under this
    License to anyone who comes into possession of a copy.  This
    License will therefore apply, along with any applicable section 7
    additional terms, to the whole of the work, and all its parts,
    regardless of how they are packaged.  This License gives no
    permission to license the work in any other way, but it does not
    invalidate such permission if you have separately received it.

    d) If the work has interactive user interfaces, each must display
    Appropriate Legal Notices; however, if the Program has interactive
    interfaces that do not display Appropriate Legal Notices, your
    work need not make them do so.

  A compilation of a covered work with other separate and independent
works, which are not by their nature extensions of the covered work,
and which are not combined with it such as to form a larger program,
in or on a volume of a storage or distribution medium, is called an
"aggregate" if the compilation and its resulting copyright are not
used to limit the access or legal rights of the compilation's users
beyond what the individual works permit.  Inclusion of a covered work
in an aggregate does not cause this License to apply to the other
parts of the aggregate.

  6. Conveying Non-Source Forms.

  You may convey a covered work in object code form under the terms
of sections 4 and 5, provided that you also convey the
machine-readable Corresponding Source under the terms of this License,
in one of these ways:

    a) Convey the object code in, or embodied in, a physical product
    (including a physical distribution medium), accompanied by the
    Corresponding Source fixed on a durable physical medium
    customarily used for software interchange.

    b) Convey the object code in, or embodied in, a physical product
    (including a physical distribution medium), accompanied by a
    written offer, valid for at least three years and valid for as
    long as you offer spare parts or customer support for that product
    model, to give anyone who possesses the object code either (1) a
    copy of the Corresponding Source for all the software in the
    product that is covered by this License, on a durable physical
    medium customarily used for software interchange, for a price no
    more than your reasonable cost of physically performing this
    conveying of source, or (2) access to copy the
    Corresponding Source from a network server at no charge.

    c) Convey individual copies of the object code with a copy of the
    written offer to provide the Corresponding Source.  This
    alternative is allowed only occasionally and noncommercially, and
    only if you received the object code with such an offer, in accord
    with subsection 6b.

    d) Convey the object code by offering access from a designated
    place (gratis or for a charge), and offer equivalent access to the
    Corresponding Source in the same way through the same place at no
    further charge.  You need not require recipients to copy the
    Corresponding Source along with the object code.  If the place to
    copy the object code is a network server, the Corresponding Source
    may be on a different server (operated by you or a third party)
    that supports equivalent copying facilities, provided you maintain
    clear directions next to the object code saying where to find the
    Corresponding Source.  Regardless of what server hosts the
    Corresponding Source, you remain obligated to ensure that it is
    available for as long as needed to satisfy these requirements.

    e) Convey the object code using peer-to-peer transmission, provided
    you inform other peers where the object code and Corresponding
    Source of the work are being offered to the general public at no
    charge under subsection 6d.

  A separable portion of the object code, whose source code is excluded
from the Corresponding Source as a System Library, need not be
included in conveying the object code work.

  A "User Product" is either (1) a "consumer product", which means any
tangible personal property which is normally used for personal, family,
or household purposes, or (2) anything designed or sold for incorporation
into a dwelling.  In determining whether a product is a consumer product,
doubtful cases shall be resolved in favor of coverage.  For a particular
product received by a particular user, "normally used" refers to a
typical or common use of that class of product, regardless of the status
of the particular user or of the way in which the particular user
actually uses, or expects or is expected to use, the product.  A product
is a consumer product regardless of whether the product has substantial
commercial, industrial or non-consumer uses, unless such uses represent
the only significant mode of use of the product.

  "Installation Information" for a User Product means any methods,
procedures, authorization keys, or other information required to install
and execute modified versions of a covered work in that User Product from
a modified version of its Corresponding Source.  The information must
suffice to ensure that the continued functioning of the modified object
code is in no case prevented or interfered with solely because
modification has been made.

  If you convey an object code work under this section in, or with, or
specifically for use in, a User Product, and the conveying occurs as
part of a transaction in which the right of possession and use of the
User Product is transferred to the recipient in perpetuity or for a
fixed term (regardless of how the transaction is characterized), the
Corresponding Source conveyed under this section must be accompanied
by the Installation Information.  But this requirement does not apply
if neither you nor any third party retains the ability to install
modified object code on the User Product (for example, the work has
been installed in ROM).

  The requirement to provide Installation Information does not include a
requirement to continue to provide support service, warranty, or updates
for a work that has been modified or installed by the recipient, or for
the User Product in which it has been modified or installed.  Access to a
network may be denied when the modification itself materially and
adversely affects the operation of the network or violates the rules and
protocols for communication across the network.

  Corresponding Source conveyed, and Installation Information provided,
in accord with this section must be in a format that is publicly
documented (and with an implementation available to the public in
source code form), and must require no special password or key for
unpacking, reading or copying.

  7. Additional Terms.

  "Additional permissions" are terms that supplement the terms of this
License by making exceptions from one or more of its conditions.
Additional permissions that are applicable to the entire Program shall
be treated as though they were included in this License, to the extent
that they are valid under applicable law.  If additional permissions
apply only to part of the Program, that part may be used separately
under those permissions, but the entire Program remains governed by
this License without regard to the additional permissions.

  When you convey a copy of a covered work, you may at your option
remove any additional permissions from that copy, or from any part of
it.  (Additional permissions may be written to require their own
removal in certain cases when you modify the work.)  You may place
additional permissions on material, added by you to a covered work,
for which you have or can give appropriate copyright permission.

  Notwithstanding any other provision of this License, for material you
add to a covered work, you may (if authorized by the copyright holders of
that material) supplement the terms of this License with terms:

    a) Disclaiming warranty or limiting liability differently from the
    terms of sections 15 and 16 of this License; or

    b) Requiring preservation of specified reasonable legal notices or
    author attributions in that material or in the Appropriate Legal
    Notices displayed by works containing it; or

    c) Prohibiting misrepresentation of the origin of that material, or
    requiring that modified versions of such material be marked in
    reasonable ways as different from the original version; or

    d) Limiting the use for publicity purposes of names of licensors or
    authors of the material; or

    e) Declining to grant rights under trademark law for use of some
    trade names, trademarks, or service marks; or

    f) Requiring indemnification of licensors and authors of that
    material by anyone who conveys the material (or modified versions of
    it) with contractual assumptions of liability to the recipient, for
    any liability that these contractual assumptions directly impose on
    those licensors and authors.

  All other non-permissive additional terms are considered "further
restrictions" within the meaning of section 10.  If the Program as you
received it, or any part of it, contains a notice stating that it is
governed by this License along with a term that is a further
restriction, you may remove that term.  If a license document contains
a further restriction but permits relicensing or conveying under this
License, you may add to a covered work material governed by the terms
of that license document, provided that the further restriction does
not survive such relicensing or conveying.

  If you add terms to a covered work in accord with this section, you
must place, in the relevant source files, a statement of the
additional terms that apply to those files, or a notice indicating
where to find the applicable terms.

  Additional terms, permissive or non-permissive, may be stated in the
form of a separately written license, or stated as exceptions;
the above requirements apply either way.

  8. Termination.

  You may not propagate or modify a covered work except as expressly
provided under this License.  Any attempt otherwise to propagate or
modify it is void, and will automatically terminate your rights under
this License (including any patent licenses granted under the third
paragraph of section 11).

  However, if you cease all violation of this License, then your
license from a particular copyright holder is reinstated (a)
provisionally, unless and until the copyright holder explicitly and
finally terminates your license, and (b) permanently, if the copyright
holder fails to notify you of the violation by some reasonable means
prior to 60 days after the cessation.

  Moreover, your license from a particular copyright holder is
reinstated permanently if the copyright holder notifies you of the
violation by some reasonable means, this is the first time you have
received notice of violation of this License (for any work) from that
copyright holder, and you cure the violation prior to 30 days after
your receipt of the notice.

  Termination of your rights under this section does not terminate the
licenses of parties who have received copies or rights from you under
this License.  If your rights have been terminated and not permanently
reinstated, you do not qualify to receive new licenses for the same
material under section 10.

  9. Acceptance Not Required for Having Copies.

  You are not required to accept this License in order to receive or
run a copy of the Program.  Ancillary propagation of a covered work
occurring solely as a consequence of using peer-to-peer transmission
to receive a copy likewise does not require acceptance.  However,
nothing other than this License grants you permission to propagate or
modify any covered work.  These actions infringe copyright if you do
not accept this License.  Therefore, by modifying or propagating a
covered work, you indicate your acceptance of this License to do so.

  10. Automatic Licensing of Downstream Recipients.

  Each time you convey a covered work, the recipient automatically
receives a license from the original licensors, to run, modify and
propagate that work, subject to this License.  You are not responsible
for enforcing compliance by third parties with this License.

  An "entity transaction" is a transaction transferring control of an
organization, or substantially all assets of one, or subdividing an
organization, or merging organizations.  If propagation of a covered
work results from an entity transaction, each party to that
transaction who receives a copy of the work also receives whatever
licenses to the work the party's predecessor in interest had or could
give under the previous paragraph, plus a right to possession of the
Corresponding Source of the work from the predecessor in interest, if
the predecessor has it or can get it with reasonable efforts.

  You may not impose any further restrictions on the exercise of the
rights granted or affirmed under this License.  For example, you may
not impose a license fee, royalty, or other charge for exercise of
rights granted under this License, and you may not initiate litigation
(including a cross-claim or counterclaim in a lawsuit) alleging that
any patent claim is infringed by making, using, selling, offering for
sale, or importing the Program or any portion of it.

  11. Patents.

  A "contributor" is a copyright holder who authorizes use under this
License of the Program or a work on which the Program is based.  The
work thus licensed is called the contributor's "contributor version".

  A contributor's "essential patent claims" are all patent claims
owned or controlled by the contributor, whether already acquired or
hereafter acquired, that would be infringed by some manner, permitted
by this License, of making, using, or selling its contributor version,
but do not include claims that would be infringed only as a
consequence of further modification of the contributor version.  For
purposes of this definition, "control" includes the right to grant
patent sublicenses in a manner consistent with the requirements of
this License.

  Each contributor grants you a non-exclusive, worldwide, royalty-free
patent license under the contributor's essential patent claims, to
make, use, sell, offer for sale, import and otherwise run, modify and
propagate the contents of its contributor version.

  In the following three paragraphs, a "patent license" is any express
agreement or commitment, however denominated, not to enforce a patent
(such as an express permission to practice a patent or covenant not to
sue for patent infringement).  To "grant" such a patent license to a
party means to make such an agreement or commitment not to enforce a
patent against the party.

  If you convey a covered work, knowingly relying on a patent license,
and the Corresponding Source of the work is not available for anyone
to copy, free of charge and under the terms of this License, through a
publicly available network server or other readily accessible means,
then you must either (1) cause the Corresponding Source to be so
available, or (2) arrange to deprive yourself of the benefit of the
patent license for this particular work, or (3) arrange, in a manner
consistent with the requirements of this License, to extend the patent
license to downstream recipients.  "Knowingly relying" means you have
actual knowledge that, but for the patent license, your conveying the
covered work in a country, or your recipient's use of the covered work
in a country, would infringe one or more identifiable patents in that
country that you have reason to believe are valid.

  If, pursuant to or in connection with a single transaction or
arrangement, you convey, or propagate by procuring conveyance of, a
covered work, and grant a patent license to some of the parties
receiving the covered work authorizing them to use, propagate, modify
or convey a specific copy of the covered work, then the patent license
you grant is automatically extended to all recipients of the covered
work and works based on it.

  A patent license is "discriminatory" if it does not include within
the scope of its coverage, prohibits the exercise of, or is
conditioned on the non-exercise of one or more of the rights that are
specifically granted under this License.  You may not convey a covered
work if you are a party to an arrangement with a third party that is
in the business of distributing software, under which you make payment
to the third party based on the extent of your activity of conveying
the work, and under which the third party grants, to any of the
parties who would receive the covered work from you, a discriminatory
patent license (a) in connection with copies of the covered work
conveyed by you (or copies made from those copies), or (b) primarily
for and in connection with specific products or compilations that
contain the covered work, unless you entered into that arrangement,
or that patent license was granted, prior to 28 March 2007.

  Nothing in this License shall be construed as excluding or limiting
any implied license or other defenses to infringement that may
otherwise be available to you under applicable patent law.

  12. No Surrender of Others' Freedom.

  If conditions are imposed on you (whether by court order, agreement or
otherwise) that contradict the conditions of this License, they do not
excuse you from the conditions of this License.  If you cannot convey a
covered work so as to satisfy simultaneously your obligations under this
License and any other pertinent obligations, then as a consequence you may
not convey it at all.  For example, if you agree to terms that obligate you
to collect a royalty for further conveying from those to whom you convey
the Program, the only way you could satisfy both those terms and this
License would be to refrain entirely from conveying the Program.

  13. Use with the GNU Affero General Public License.

  Notwithstanding any other provision of this License, you have
permission to link or combine any covered work with a work licensed
under version 3 of the GNU Affero General Public License into a single
combined work, and to convey the resulting work.  The terms of this
License will continue to apply to the part which is the covered work,
but the special requirements of the GNU Affero General Public License,
section 13, concerning interaction through a network will apply to the
combination as such.

  14. Revised Versions of this License.

  The Free Software Foundation may publish revised and/or new versions of
the GNU General Public License from time to time.  Such new versions will
be similar in spirit to the present version, but may differ in detail to
address new problems or concerns.

  Each version is given a distinguishing version number.  If the
Program specifies that a certain numbered version of the GNU General
Public License "or any later version" applies to it, you have the
option of following the terms and conditions either of that numbered
version or of any later version published by the Free Software
Foundation.  If the Program does not specify a version number of the
GNU General Public License, you may choose any version ever published
by the Free Software Foundation.

  If the Program specifies that a proxy can decide which future
versions of the GNU General Public License can be used, that proxy's
public statement of acceptance of a version permanently authorizes you
to choose that version for the Program.

  Later license versions may give you additional or different
permissions.  However, no additional obligations are imposed on any
author or copyright holder as a result of your choosing to follow a
later version.

  15. Disclaimer of Warranty.

  THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

  16. Limitation of Liability.

  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.

  17. Interpretation of Sections 15 and 16.

  If the disclaimer of warranty and limitation of liability provided
above cannot be given local legal effect according to their terms,
reviewing courts shall apply local law that most closely approximates
an absolute waiver of all civil liability in connection with the
Program, unless a warranty or assumption of liability accompanies a
copy of the Program in return for a fee.

                     END OF TERMS AND CONDITIONS

            How to Apply These Terms to Your New Programs

  If you develop a new program, and you want it to be of the greatest
possible use to the public, the best way to achieve this is to make it
free software which everyone can redistribute and change under these terms.

  To do so, attach the following notices to the program.  It is safest
to attach them to the start of each source file to most effectively
state the exclusion of warranty; and each file should have at least
the "copyright" line and a pointer to where the full notice is found.

    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) <year>  <name of author>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

Also add information on how to contact you by electronic and paper mail.

  If the program does terminal interaction, make it output a short
notice like this when it starts in an interactive mode:

    <program>  Copyright (C) <year>  <name of author>
    This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
    This is free software, and you are welcome to redistribute it
    under certain conditions; type `show c' for details.

The hypothetical commands `show w' and `show c' should show the appropriate
parts of the General Public License.  Of course, your program's commands
might be different; for a GUI interface, you would use an "about box".

  You should also get your employer (if you work as a programmer) or school,
if any, to sign a "copyright disclaimer" for the program, if necessary.
For more information on this, and how to apply and follow the GNU GPL, see
<https://www.gnu.org/licenses/>.

  The GNU General Public License does not permit incorporating your program
into proprietary programs.  If your program is a subroutine library, you
may consider it more useful to permit linking proprietary applications with
the library.  If this is what you want to do, use the GNU Lesser General
Public License instead of this License.  But first, please read
<https://www.gnu.org/licenses/why-not-lgpl.html>.****
"""
import asyncio
import json
import os
import sys
from pathlib import Path
from typing import Optional

import dotenv
import sqlalchemy
import typer
from click import MissingParameter, UsageError
from dotenv import dotenv_values, load_dotenv
from .engine_and_database import (
    add_products_to_db,
    add_suppliers_to_db,
    create_db_engine,
    save_all_changes,
)
from .info_message import update_db_success_sqlite, update_db_with_success
from loguru import logger
from rich import print as rprint
from .scrape_from_disk import PageParser
from sqlmodel import SQLModel
from typing_extensions import Annotated
from .web_scrapper import async_scrapper, sync_scrapper
from . import LOGURU_LEVEL

load_dotenv()
logger.remove(0)
logger.add(sys.stderr, colorize=True, level=f"{LOGURU_LEVEL}")  # type: ignore

app = typer.Typer(pretty_exceptions_enable=False)


def _db_url(credentials: dict = dict(), auto_fill: bool = False):
    if auto_fill is True:
        if Path("db_credentials.json").exists() is False:
            raise FileNotFoundError(
                "db_credentials.json is missing YOU NEED TO RUN `db-init` COMMAND FIRST !!"
            )
        with open("db_credentials.json", "r") as f:
            cred: dict = json.load(f)
            items = credentials.items()
            for item in items:
                if item[0] == "host" and item[1] not in ["localhost", None]:
                    cred.update({"host": item[1]})
                elif item[0] == "port" and item[1] not in [3306, None]:
                    cred.update({"port": item[1]})
                elif item[0] == "db_name" and item[1] not in [
                    cred.get("db_name"),
                    None,
                ]:
                    cred.update({"db_name": item[1]})
                elif item[0] == "user" and item[1] not in [cred.get("user"), None]:
                    cred.update({"user": item[1]})
                elif item[0] == "password" and item[1] not in [
                    cred.get("password"),
                    None,
                ]:
                    cred.update({"password": item[1]})
                else:
                    continue
        return f"mysql+mysqldb://{cred.get('user')}:{cred.get('password')}@{cred.get('host')}/{cred.get('db_name')}"

    else:
        cred = {
            "user": credentials.get("user"),
            "password": credentials.get("password"),
            "host": credentials.get("host"),
            "port": credentials.get("port"),
            "db_name": credentials.get("db_name"),
        }
        for key, value in cred.items():
            if value is None:
                raise MissingParameter(f"{key} is missing")
        with open("db_credentials.json", "w") as f:
            json.dump(cred, f)
        return f"mysql+mysqldb://{cred.get('user')}:{cred.get('password')}@{cred.get('host')}/{cred.get('db_name')}"


@app.command()
def scraper(
    key_words: Annotated[str, typer.Argument(help="Keywords to search for on alibaba")] = "",
    html_folder: Annotated[
        Optional[str], typer.Option("--html-folder","-hf",help="Folder to save the results")
    ] = None,
    sync_api: Annotated[
        Optional[bool], typer.Option("--sync-api","-sa",help="wether to sync or not")
    ] = False,
    page_results: Annotated[
        int, typer.Option("--page-results","-pr",help="Number of results per page to scrape from alibaba 10 by defaults")
    ] = 10,
) -> None:
    """
    Runs the web scrapper and loking for products and related suppliers infos based on provided keywords, then save the results in a specified folder by default this folder is named with the keywords that was provided and could attached with '_' if there are spaces in the keywords'.
    """
    save_in_folder = (
        key_words.strip().replace(" ", "_")
        if (html_folder is None and html_folder != "")
        else html_folder
    )
    if sync_api:
        sync_scrapper(save_in=save_in_folder, key_words=key_words, page_results=page_results)
    else:
        asyncio.run(async_scrapper(save_in=save_in_folder, key_words=key_words, page_results=page_results))


@app.command()
def db_update(
    kw_results:Annotated[Path, typer.Option("--kw-results","-kr",help="Folder where the html results are stored")],
    db_engine: Annotated[
        Optional[str], typer.Argument(help="Name of database engine to use")
    ] = "sqlite",
    filename: Annotated[
        Optional[str],
        typer.Option(
            "--filename","-f",help="Name of the sqlite file(without any extensions) to update with news data",
        ),
    ] = None,
    host: Annotated[
        Optional[str],
        typer.Option(
            "--host","-h",help="Host of the database engine",
        ),
    ] = "localhost",
    port: Annotated[
        Optional[int], typer.Option("--port","-p",help="Port of the database engine")
    ] = 3306,
    user: Annotated[
        Optional[str],
        typer.Option("--user","-u",help="User of the database engine", show_default=False),
    ] = None,
    password: Annotated[
        Optional[str],
        typer.Option("--password","-pw",help="Password of the database engine", show_default=False),
    ] = None,
    db_name: Annotated[
        Optional[str],
        typer.Option("--db-name","-db",help="Database of the database engine", show_default=False),
    ] = None,
):
    """
    Updates the database with new products and their related suppliers. \n
    Example: \n
    \b \b aba db-update sqlite/mysql --kw-results keyword_folder_result --db-name your_database_name \n 
    Raises: \n
    \b \b \b MissingParameter: If the engine is not 'sqlite' or 'mysql', or if the filename is required for sqlite engine, or if the filename is specified for non-sqlite engine.

    Returns:
        None
    """
    if db_engine not in ["sqlite", "mysql"]:
        raise MissingParameter("--engine should be sqlite or mysql")
    if db_engine == "sqlite" and filename is None:
        raise MissingParameter("--filename is required for sqlite engine")
    if db_engine != "sqlite" and filename is not None:
        raise MissingParameter(
            "You don't need to specify --filename for non-sqlite engine"
        )
    if db_engine == "mysql":
        db_url = _db_url(credentials=locals(), auto_fill=True)
        mysql_engine = create_db_engine(db_url=db_url)
        page_parser = PageParser(targeted_folder=kw_results.resolve())
        suppliers = page_parser.detected_suppliers()
        products = page_parser.detected_products()
        add_suppliers_to_db(suppliers=suppliers, engine_db=mysql_engine)
        add_products_to_db(products=products, engine_db=mysql_engine)
        save_all_changes(engine_db=mysql_engine, sql_model=SQLModel)
        update_db_with_success()

    if not kw_results.exists():
        raise UsageError(f"Folder {kw_results} does not exist")        
    else:
        page_parser = PageParser(targeted_folder=kw_results)
        suppliers = page_parser.detected_suppliers()
        products = page_parser.detected_products()
        sqlite_engine = create_db_engine(db_name=filename)  # type: ignore
        add_suppliers_to_db(suppliers=suppliers, engine_db=sqlite_engine)
        add_products_to_db(products=products, engine_db=sqlite_engine)
        save_all_changes(engine_db=sqlite_engine, sql_model=SQLModel)
        update_db_success_sqlite(sqlite_file=filename)  # type: ignore

# TODO : add message when database initialization end with succes or when sommething went wrong with throught this process
@app.command()
def db_init(
    engine: Annotated[
        str, typer.Argument(help="Name of database engine to use")
    ] = "sqlite",
    sqlite_file: Annotated[
        Optional[str],
        typer.Option("--sqlite-file","-f",
            help="Name of the sqlite file to use (without any extensions)",
            show_default=False,
        ),
    ] = None,
    host: Annotated[
        Optional[str],
        typer.Option(
            "--host","-h",help="Host of the database engine",
        ),
    ] = "localhost",
    port: Annotated[
        Optional[int], typer.Option("--port","-p",help="Port of the database engine")
    ] = 3306,
    user: Annotated[
        Optional[str],
        typer.Option("--user","-u",help="User of the database engine"),
    ] = None,
    password: Annotated[
        Optional[str],
        typer.Option("--password","-pw",help="Password of the database engine"),
    ] = None,
    db_name: Annotated[
        Optional[str],
        typer.Option("--db-name","-db",help="Database of the database engine", ),
    ] = None,
    only_with: Annotated[
        Optional[bool],
        typer.Option(
            "--only-with","-ow",help="set it to true if you just want to update some database crendentials but not all",
        ),
    ] = False
):
    """
    This command initializes a database using the specified engine. It checks the validity of the engine and the
    required parameters. If the engine is 'sqlite', it creates a sqlite database file with the specified name. If the
    engine is 'mysql', it creates a mysql database using the specified host, port, user, password, and database name provided by the user. \n
    If your database has been initialized with mysql engine, a json file will be created with your credentials. Making it easier to update your database later means you will no longer need to write all your credentials each time you want to update.

    Raises:\n
    \b \b \b MissingParameter: If the engine is not 'sqlite' or 'mysql', or if the filename is required for sqlite engine, or if the filename is specified for non-sqlite engine.


    """
    if engine not in ["sqlite", "mysql"]:
        raise MissingParameter("--engine should be sqlite or mysql")
    if engine == "sqlite" and sqlite_file is None:
        raise MissingParameter("--sqlite-file is required for sqlite engine")
    if engine != "sqlite" and sqlite_file is not None:
        raise MissingParameter(
            "You dont need to specify --sqlite-file for non-sqlite engine"
        )
    if engine == "mysql":
        if only_with:
            db_url = _db_url(credentials=locals(), auto_fill=True)
        else:
            db_url = _db_url(credentials=locals())
        mysql_engine = create_db_engine(db_url=db_url)
        try:
            save_all_changes(engine_db=mysql_engine, sql_model=SQLModel)
            rprint(
                f"[bold white]Database [magenta bold] {db_name} [/magenta bold] has been created succesfully :white_heavy_check_mark-emoji: ![/bold white]"
            )
        except sqlalchemy.exc.OperationalError as e:
            if "Unknown database" in str(e):
                rprint(
                    f"[bold white blink] It's seems like you want to use mysql engine but you haven't created database: [magenta bold] `{db_name}` [/magenta bold] ![/bold white blink] create it first and then try again"
                )
            logger.error(f"Errors has occured: {e}")
    else:
        sqlite_engine = create_db_engine(db_name=sqlite_file)  # type: ignore
        save_all_changes(engine_db=sqlite_engine, sql_model=SQLModel)
        rprint(
            f"[bold white]Database [magenta bold] {sqlite_file}.sqlite [/magenta bold] has been created succesfully :white_heavy_check_mark-emoji: ![/bold white]"
        )






import asyncio
import json
import os
import sys
from pathlib import Path
from typing import Optional

import sqlalchemy
import typer
from click import MissingParameter
from dotenv import load_dotenv
from .engine_and_database import (
    add_products_to_db,
    add_suppliers_to_db,
    create_db_engine,
    save_all_changes,
)
from .info_message import update_db_success_sqlite, update_db_with_success
from loguru import logger
from rich import print as rprint
from .scrape_from_disk import PageParser
from sqlmodel import SQLModel
from typing_extensions import Annotated
from .web_scrapper import async_scrapper, sync_scrapper

load_dotenv()

logger.remove(0)
logger.add(sys.stderr, colorize=True, level=os.environ.get("LOGURU_LEVEL", ""))

app = typer.Typer(pretty_exceptions_enable=False)


def _db_url(credentials: dict = dict(), auto_fill: bool = False):
    if auto_fill is True:
        if Path("db_credentials.json").exists() is False:
            raise FileNotFoundError(
                "db_credentials.json is missing YOU NEED TO RUN `db-init` COMMAND FIRST !!"
            )
        with open("db_credentials.json", "r") as f:
            cred: dict = json.load(f)
            items = credentials.items()
            for item in items:
                if item[0] == "host" and item[1] not in ["localhost", None]:
                    cred.update({"host": item[1]})
                elif item[0] == "port" and item[1] not in [3306, None]:
                    cred.update({"port": item[1]})
                elif item[0] == "db_name" and item[1] not in [
                    cred.get("db_name"),
                    None,
                ]:
                    cred.update({"db_name": item[1]})
                elif item[0] == "user" and item[1] not in [cred.get("user"), None]:
                    cred.update({"user": item[1]})
                elif item[0] == "password" and item[1] not in [
                    cred.get("password"),
                    None,
                ]:
                    cred.update({"password": item[1]})
                else:
                    continue
        return f"mysql+mysqldb://{cred.get('user')}:{cred.get('password')}@{cred.get('host')}/{cred.get('db_name')}"

    else:
        cred = {
            "user": credentials.get("user"),
            "password": credentials.get("password"),
            "host": credentials.get("host"),
            "port": credentials.get("port"),
            "db_name": credentials.get("db_name"),
        }
        for key, value in cred.items():
            if value is None:
                raise MissingParameter(f"{key} is missing")
        with open("db_credentials.json", "w") as f:
            json.dump(cred, f)
        return f"mysql+mysqldb://{cred.get('user')}:{cred.get('password')}@{cred.get('host')}/{cred.get('db_name')}"


@app.command()
def run_scrapper(
    key_words: Annotated[str, typer.Argument(help="Keywords to search")],
    html_folder: Annotated[
        Optional[str], typer.Option(help="Folder to save the results")
    ] = None,
    sync_api: Annotated[
        Optional[bool], typer.Option(help="wether to sync or not",default=False)
    ] = False,
) -> None:
    """
    Runs the web scrapper and loking for products and related suppliers infos based on provided keywords, then save the results in a specified folder by default this folder is named with the keywords that was provided and could attached with '_' if there are spaces in the keywords'.

    Args:
        key_words (str): The keywords to search for.
        html_folder (Optional[str], optional): The folder to save the results. Defaults to None.

    Returns:
        None

    Raises:
        None
    """

    save_in_folder = (
        key_words.strip().replace(" ", "_")
        if (html_folder is None and html_folder != "")
        else html_folder
    )
    if sync_api:
        sync_scrapper(save_in=save_in_folder, key_words=key_words)
    else:
        asyncio.run(async_scrapper(save_in=save_in_folder, key_words=key_words))


@app.command()
def db_update(
    db_engine: Annotated[
        Optional[str], typer.Argument(help="Name of database engine to use")
    ] = "sqlite",
    kw_results=typer.Option(
        default=..., help="Folder where the html results are stored"
    ),
    filename: Annotated[
        Optional[str],
        typer.Option(
            help="Name of the sqlite file(without any extensions) to update with news data",
        ),
    ] = None,
    host: Annotated[
        Optional[str],
        typer.Option(
            help="Host of the database engine",
        ),
    ] = "localhost",
    port: Annotated[
        Optional[int], typer.Option(help="Port of the database engine")
    ] = 3306,
    user: Annotated[
        Optional[str],
        typer.Option(help="User of the database engine", show_default=False),
    ] = None,
    password: Annotated[
        Optional[str],
        typer.Option(help="Password of the database engine", show_default=False),
    ] = None,
    db_name: Annotated[
        Optional[str],
        typer.Option(help="Database of the database engine", show_default=False),
    ] = None,
):
    """
    Updates the database with new products and their related suppliers.
    Raises:
        MissingParameter: If the engine is not 'sqlite' or 'mysql', or if the filename is required for sqlite engine, or if the filename is specified for non-sqlite engine.

    Returns:
        None
    """
    if db_engine not in ["sqlite", "mysql"]:
        raise MissingParameter("--engine should be sqlite or mysql")
    if db_engine == "sqlite" and filename is None:
        raise MissingParameter("--filename is required for sqlite engine")
    if db_engine != "sqlite" and filename is not None:
        raise MissingParameter(
            "You don't need to specify --filename for non-sqlite engine"
        )
    if db_engine == "mysql":
        db_url = _db_url(credentials=locals(), auto_fill=True)
        mysql_engine = create_db_engine(db_url=db_url)
        page_parser = PageParser(targeted_folder=kw_results)
        suppliers = page_parser.detected_suppliers()
        products = page_parser.detected_products()
        add_suppliers_to_db(suppliers=suppliers, engine_db=mysql_engine)
        add_products_to_db(products=products, engine_db=mysql_engine)
        save_all_changes(engine_db=mysql_engine, sql_model=SQLModel)
        update_db_with_success()

    else:
        page_parser = PageParser(targeted_folder=kw_results)
        suppliers = page_parser.detected_suppliers()
        products = page_parser.detected_products()
        sqlite_engine = create_db_engine(db_name=filename)  # type: ignore
        add_suppliers_to_db(suppliers=suppliers, engine_db=sqlite_engine)
        add_products_to_db(products=products, engine_db=sqlite_engine)
        save_all_changes(engine_db=sqlite_engine, sql_model=SQLModel)
        update_db_success_sqlite(sqlite_file=filename)  # type: ignore


@app.command()
def db_init(
    engine: Annotated[
        str, typer.Argument(help="Name of database engine to use")
    ] = "sqlite",
    sqlite_file: Annotated[
        Optional[str],
        typer.Option(
            help="Name of the sqlite file to use (without any extensions)",
            show_default=False,
        ),
    ] = None,
    host: Annotated[
        Optional[str],
        typer.Option(
            help="Host of the database engine",
        ),
    ] = "localhost",
    port: Annotated[
        Optional[int], typer.Option(help="Port of the database engine")
    ] = 3306,
    user: Annotated[
        Optional[str],
        typer.Option(help="User of the database engine"),
    ] = None,
    password: Annotated[
        Optional[str],
        typer.Option(help="Password of the database engine"),
    ] = None,
    db_name: Annotated[
        Optional[str],
        typer.Option(help="Database of the database engine", ),
    ] = None,
    only_with: Annotated[
        Optional[bool],
        typer.Option(
            help="set it to true if you just want to update some database crendentials but not all",
        ),
    ] = False
):
    """
    Initializes a database using the specified engine.

    Raises:
        MissingParameter: If the engine is not 'sqlite' or 'mysql', or if the filename is required for sqlite engine,
            or if the filename is specified for non-sqlite engine.


    This function initializes a database using the specified engine. It checks the validity of the engine and the
    required parameters. If the engine is 'sqlite', it creates a sqlite database file with the specified name. If the
    engine is 'mysql', it creates a mysql database using the specified host, port, user, password, and database name.
    After creating the database, it saves the changes to the database. Finally,
    it prints a success message indicating the creation of the database.
    """
    if engine not in ["sqlite", "mysql"]:
        raise MissingParameter("--engine should be sqlite or mysql")
    if engine == "sqlite" and sqlite_file is None:
        raise MissingParameter("--sqlite-file is required for sqlite engine")
    if engine != "sqlite" and sqlite_file is not None:
        raise MissingParameter(
            "You dont need to specify --sqlite-file for non-sqlite engine"
        )
    if engine == "mysql":
        if only_with:
            db_url = _db_url(credentials=locals(), auto_fill=True)
        else:
            db_url = _db_url(credentials=locals())
        mysql_engine = create_db_engine(db_url=db_url)
        try:
            save_all_changes(engine_db=mysql_engine, sql_model=SQLModel)
            rprint(
                f"[bold white]Database [magenta bold] {db_name} [/magenta bold] has been created succesfully :white_heavy_check_mark-emoji: ![/bold white]"
            )
        except sqlalchemy.exc.OperationalError as e:
            if "Unknown database" in str(e):
                rprint(
                    f"[bold white blink] It's seems like you want to use mysql engine but you haven't created database: [magenta bold] `{db_name}` [/magenta bold] ![/bold white blink] create it first and then try again"
                )
            logger.error(f"Errors has occured: {e}")
    else:
        sqlite_engine = create_db_engine(db_name=sqlite_file)  # type: ignore
        save_all_changes(engine_db=sqlite_engine, sql_model=SQLModel)






from asyncio import Task
import html
from pathlib import Path
import json 

import selectolax
from loguru import logger


def _create_folder(folder_name: str):
    logger.info(f"Create folder with {folder_name} as name if not exists yet ...")
    Path(f"{folder_name}").mkdir(exist_ok=True)
    return Path(f"{folder_name}")


def json_parser_to_dict(html_content: str | bytes,css_selector: str = ""):
    """Parse the HTML content of the page to get the divs with class `.organic-list.viewtype-list`

    :param html_content: The HTML content of the page
    :type html_content: str
    :return: A list of divs with class `.organic-list.viewtype-list`
    :rtype: list
    """

    # dont parse anything at this stage as we are going to parse later
    # print(html_content)
    body_parser = selectolax.parser.HTMLParser(html_content)
    script_div = body_parser.css_first(css_selector)
    if script_div is None:
        logger.warning(
            "any HTML content  match the selector 'body > div.container > script:nth-child(9)', None will be returned and you may not have enough data as expected"
        )
        return None
    json_result = script_div.text().replace("window.__page__data__config =","").replace("window.__page__data = window.__page__data__config.props","").strip("\n \t \b ")
    if json_result is None:
        logger.warning(
            "any HTML content  match the selector '.organic-list', None value has been returned"
        )
        raise ValueError("None value has been returned")
    
    elif "window.__icbusearch_layout_i18n_kv__" in json_result:
        json_result = json_parser_to_dict(html_content=html_content, css_selector="body > div.container > script:nth-child(9)")
        return json_result
    else:
        # print(json_result)
        return json_result.strip()


@logger.catch()
def write_to_disk(folder_name: str, pages_contents: list[str]):
    """
    Writes list's contents HTML pages to disk.

    Args:
        folder_name (str): The name of the folder where the HTML files will be saved.
        pages_contents (list[str]): A list of strings representing the contents of the HTML pages.

    Returns:
        None

    Raises:
        None

    Side Effects:
        - Creates a folder with the given name if it does not exist.
        - Saves each HTML page as a separate file in the folder.
        - Logs the progress of saving each HTML page.

    Example Usage:
        write_to_disk("my_folder", ["<html>...</html>", "<html>...</html>"])
    """

    logger.info("Saving your scrapping result in disk...")
    path_obj = _create_folder(folder_name)
    for page_number, content in enumerate(pages_contents):
        # parsed_content = json_parser_to_dict(html_content=content, css_selector="body > div.container > script:nth-child(8)")
        # # print(parsed_content)
        # if parsed_content is None:
        #     logger.warning(
        #         f"parsed content is none"
        #     )
        #     continue
        # json_as_dict = json.loads(parsed_content)
        # print(json_as_dict)
        # print(type(json_as_dict))
        # json_as_str = json.dumps(json_as_dict, indent=4)
        with open(
            (path_obj / f"page_{page_number+1}.html").resolve(), "w", encoding="utf-8"
        ) as file:
            file.write(content)
        logger.info(
            f"Html content from page {page_number+1} has been saved succesfully !"
        )


if __name__ == "__name__":
    ...



import sqlite3
from typing import Any, Sequence

from click import UsageError
from loguru import logger
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn, TimeElapsedColumn, SpinnerColumn

from aba_cli_scrapper.typed_datas import ProductDict, SupplierDict
from .models import Product, Supplier  # noqa: F401
from MySQLdb import OperationalError as MySQLdbOperationalError
from sqlalchemy.engine import Engine
from sqlmodel import Session, SQLModel, create_engine, select  # noqa: F401
from sqlalchemy.exc import IntegrityError, OperationalError


def create_db_engine(db_name: str = "", db_url: str = ""):
    if db_url == "":
        engine = create_engine(f"sqlite:///{db_name}.sqlite")
    if db_url != "":
        engine = create_engine(db_url)
    return engine


# engine = create_engine("sqlite:///iphones_xs.sqlite")
def save_all_changes(
    engine_db: Engine,
    sql_model,
):
    """
    A function that saves all changes by creating all metadata for a given SQL model on the specified engine database.

    Parameters:
    engine_db (Engine): The database engine to use for saving changes.
    sql_model: The SQL model for which to create the metadata.
    """
    logger.info("Saving all changes ...")
    try:
        sql_model.metadata.create_all(engine_db)
    except MySQLdbOperationalError as e:
        logger.error(f"Errors has occured: {e}")
        raise UsageError(f"Something went wrong an unexpected error has occured:{e}") from e



def add_suppliers_to_db(suppliers: Sequence[SupplierDict], engine_db: Engine):
    """
    Adds a list of suppliers to the database.

    Args:
        suppliers (list[dict[str, Any]]): A list of dictionaries representing suppliers.
        engine_db (Engine): The SQLAlchemy engine object representing the database connection.

    Returns:
        None

    Raises:
        None
    """
    logger.info("Adding suppliers to database ...")
    with Session(engine_db) as session:
        added = set()
        with Progress(TextColumn("[progress.description]{task.description}"), BarColumn(),TaskProgressColumn(), TimeElapsedColumn(),transient=True) as progress:
            adder_supp = progress.add_task(description="Adding suppliers to database ...")
            for supplier in suppliers:
                # pprint(f"country_name={supplier['country_name']}")
                if supplier["name"] in added:
                    continue
                session.add(
                    Supplier(
                        name=supplier["name"],
                        verification_mode=supplier["verified_type"],
                        sopi_level=supplier["sopi_level"],
                        country_name=supplier["country_name"],
                        years_as_gold_supplier=supplier["gold_supplier_year"],
                        supplier_service_score= supplier["supplier_service_score"]
                    )
                )
                added.add(supplier["name"])
                try:
                    session.commit()
                    progress.update(adder_supp, advance=100/len(suppliers))
                except OperationalError as e:
                    if "no such table" in str(e):
                        raise UsageError("Your database must be initialized before if you want to update it. instead run :  `aba-run db-init sqlite sqlite-file <your_sqlite_db_file_name>`") from e
                    raise UsageError(f"Something went wrong an unexpected error has occured:{e}") from e
                except IntegrityError as e:
                    raise UsageError("Seems like you are trying to update a database which has already been updated ! Instead initialize a new one and update it with: ~aba-run db-update sqlite sqlite-file <new_sqlite_db_file_name>~") from e


def add_products_to_db(products: Sequence[ProductDict], engine_db: Engine):
    """
    Adds a list of products to the database.

    Args:
        products (list[dict[str, Any]]): A list of dictionaries representing products.
        engine_db (Engine): The SQLAlchemy engine object representing the database connection.

    Returns:
        None

    Raises:
        None
    """
    logger.info("Adding products to database ...")
    with Session(engine_db) as session:
        added = set()
        with Progress(SpinnerColumn(),TextColumn("[progress.description]{task.description}"), BarColumn(),TaskProgressColumn(), TimeElapsedColumn(),transient=True) as progress:
            try:
                add_prods = progress.add_task(description="Adding products to database ...")
                for product in products:
                    if product["name"] in added:
                        continue
                    supplier = _related_supplier(session, supplied_by=product["supplied_by"])
                    session.add(
                        Product(
                            name=product["name"],
                            alibaba_guranteed=product["guaranteed_by_alibaba"],
                            certifications=product["certifications"],
                            minimum_to_order=product["minimum_to_order"],
                            ordered_or_sold=product["ordered_or_sold"],
                            supplier_id=supplier.id,
                            min_price=product["min_price"],
                            max_price=product["max_price"],
                            trade_product= product['trade_product'],
                            review_count= product['review_count'],
                            review_score= product['review_score'],
                            product_score= product['product_score'],
                            shipping_time_score= product['shipping_time_score'],
                            is_instant_order=product["instant_order"],
                            is_customizable= product["customizable"],
                            is_full_promotion=product['is_full_promotion']


                        )
                    )
                    added.add(product["name"])
                    session.commit()
                    progress.update(add_prods, advance=100/len(products))
            except OperationalError as e:
                logger.error(f"Errors has occured: {e}")
                if "no such table" in str(e):
                    raise UsageError("Your database must be initialized first run : aba db-init --help to know more ") from e
                raise UsageError(f"Something went wrong an unexpected error has occured:{e}") from e


def _related_supplier(current_session: Session, supplied_by: str):
    logger.info("Getting related supplier ...")
    query_statement = select(Supplier).where(Supplier.name == supplied_by)
    return current_session.exec(query_statement).all()[0]


if __name__ == "__main__":
    ...



from typing import Optional

from sqlmodel import Field, SQLModel


class Product(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str = Field(index=True,unique=True)
    alibaba_guranteed: bool
    certifications: str
    minimum_to_order: int
    ordered_or_sold: int
    supplier_id: Optional[int] = Field(default=None, foreign_key="supplier.id")
    min_price: float
    max_price: float
    product_score: float
    review_count : float
    review_score : float
    shipping_time_score : float
    is_full_promotion: bool
    is_customizable: bool
    is_instant_order: bool
    trade_product:bool

class Supplier(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    name: str = Field(index=True,unique=True)
    verification_mode: str
    sopi_level: int
    country_name: str
    years_as_gold_supplier: int
    supplier_service_score: float


if __name__ == "__main__":
    ...
    # SQLModel.metadata.create_all(engine)



from rich import print as rprint


def run_scrapper_with_success(folder_name: str):
    rprint(
        "[bold white]Step :one:  [bold green] Done  :star::white_heavy_check_mark-emoji: ! [/bold green] [/bold white]  "
    )
    # rprint(
    #     "[bold white ]Scraping step has been completed [bold green]with success :white_heavy_check_mark-emoji: ![/bold green] [/bold white]"
    # )
#     rprint("""[bold white]Here we go to the step :two:   [/bold white]
#            ---
# \b\b\b initialize your database with [bold magenta blink]aba-run db-init  mysql --user your_database_username --password your_database_password --db-name alibaba_products[/bold magenta blink] command(for mysql engine)
#             or 
# \b\b\b [bold magenta blink]aba-run db-init --engine sqlite --sqlite-file sqlite_file_name(without any extensions) [/bold magenta blink] command(for sqlite engine)...""")
    # rprint(
    #     f"""[bold white ]Now all pages results matching your keywords has been saved in [magenta]{folder_name}[/magenta] now if is still not the case you must to initialize your database with [bold magenta blink]
    #     `python src/app.py db-init --engine mysql --user your_username --password your_password --db-name alibaba_products`[/bold magenta blink] command(for mysql engine).
    #     Or `python src/app.py db-init --engine sqlite --sqlite-file sqlite_file_name(without any extensions)` command(for sqlite engine). As soon as you have done that, you can update all thoses new scraped products and suppliers inside your database with [bold magenta blink]`python src/app.py db-update --db-engine mysql --kw-results {folder_name} command (for sqlite engine)`[/bold magenta blink] or [bold magenta blink]`python src/app.py db-update --db-engine sqlite --kw-results bike_results --filename {folder_name}`[/bold magenta blink]..."""
    # )


def update_db_with_success():
    rprint(
        "[bold white]Step :three:  is [bold green] done  :star: ! [/bold green] [/bold white]"
    )
    rprint(
        "[bold white blink]Your database has been updated with new products and their related suppliers :white_heavy_check_mark-emoji: ![/bold white blink]"
    )
    rprint(
        "[bold white]Now you can do some amazing analysis with your database for making your business more profitable :star: :star: ! [/bold white ]"
    )


def update_db_success_sqlite(sqlite_file: str):
    rprint(
        f"[bold white][magenta bold]{sqlite_file}.sqlite [/magenta bold] file has been updated succesfully  with new products and their related suppliers :white_heavy_check_mark-emoji: ![/bold white]"
    )
    rprint(
        "[bold white]Step :three:  is [bold green] done  :star: ! [/bold green] [/bold white]  "
    )
    rprint(
        "[bold white ]Now you can do some amazing analysis with your database for making your business more profitable :star: :star: ! [/bold white ]"
    )



"""
This Module will gather all datas from html files and build database with two tables, Products and Suppliers

"""

from collections import OrderedDict
from dataclasses import dataclass
import json
from pathlib import Path
from pprint import pprint
from typing import Sequence, Union

import rich
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn, TimeElapsedColumn,SpinnerColumn
from rich.text import Text
from loguru import logger
from selectolax.parser import HTMLParser

from aba_cli_scrapper.typed_datas import ProductDict, SupplierDict
from .utils_scrapping import (
    country_name,
    get_product_certification,
    get_product_price,
    is_alibaba_guaranteed,
    is_customizable,
    is_full_promotion,
    is_instant_order,
    is_trade_product,
    ordered_or_sold,
    suppliers_status,
)
from .html_to_disk import json_parser_to_dict

# logger.remove(4)
# logger.add(sys.stderr,level="DEBUG",colorize=True)


@dataclass
class PageParser:
    """Parse each html page downloaded from alibaba and return a list of suppliers and products

    :raises TypeError: _description_
    :raises FileNotFoundError: if the targeted folder does not exist
    :return: PageParser object
    :rtype: class `PageParser`
    """

    targeted_folder: Union[
        str, Path
    ]  # folder path that contains all html files to be scraped

    def _retrieve_html_content_as_string(self, html_file: Path):
        logger.info(f"Retrieving html content from file : {html_file}")

        if not isinstance(html_file, Path):
            raise TypeError(
                f"html_file must be a Path object but got : {type(html_file)}"
            )

        with open(html_file, "r", encoding="utf-8") as fs:
            html_content = fs.read()
        logger.info("Html content has been retrieved.")
        return html_content
    
    def _retrieve_json_content_as_dict(self, json_file: Path):
        logger.info(f"Retrieving json content from file : {json_file.resolve()}")


        if not isinstance(json_file, Path):
            raise TypeError(
                f"json_file argument must be a Path object but got : {type(json_file)}"
            )

        with open(json_file, "r", encoding="utf-8") as fs:
            json_content_as_dict = json.load(fs,)
        logger.info("Html content has been retrieved.")
        return json_content_as_dict['props']['offerResultData']['offers']

    @logger.catch(FileNotFoundError)
    def _html_files_explorer(self):
        targeted_folder = Path(self.targeted_folder).resolve()
        if not targeted_folder.exists():
            raise FileNotFoundError()
        logger.info(f"getting html files from {targeted_folder} ... ")
        html_files = [html_file for html_file in targeted_folder.glob("*.html")]
        logger.info("Html files list has been returned.")
        return html_files
    
    def _json_files_explorer(self):
        targeted_folder = Path(self.targeted_folder).resolve()
        if not targeted_folder.exists():
            raise FileNotFoundError()
        logger.info(f"getting html files from {targeted_folder} ... ")
        json_files = [html_file for html_file in targeted_folder.glob("*.json")]
        logger.info("json files list has been returned.")
        return json_files

    def _divs_and_dict(self, selector: str = ".fy23-search-card"):
        """
        Retrieve div tags with a specified selector class from HTML files, parse them, and return a list of tuples containing the parsed div tags and corresponding JSON data as dictionaries.
        
        :param selector: The CSS selector for the div tags to retrieve. Defaults to ".fy23-search-card".
        :return: A list of tuples where each tuple contains the parsed div tags and the corresponding JSON data as dictionaries.
        :rtype: List[Tuple[Node, Dict]]
        """
        logger.info(f"Retrieving  div tags with class='{selector}' ...")
        divs_and_dict = [
            (
            HTMLParser(self._retrieve_html_content_as_string(html_file)).css(selector),
            json_parser_to_dict(self._retrieve_html_content_as_string(html_file), css_selector="body > div.container > script[type='text/javascript'] + script"),
                )
            for html_file in self._html_files_explorer() 
        ]
        logger.info("All expected divs and dict has been retrived ...")
        # TODO: add good items directly in a new list instead of removing them
        _ = [divs_and_dict.pop(divs_and_dict.index(item)) for item in divs_and_dict if item[0] is not None and item[1] is None] # remove none tuple from list
        new_divs_and_dict = []
        for item in divs_and_dict:
            # print("hello")
            # print(item[1])
            # print(type(item[1]))
            new_divs_and_dict.append((item[0],json.loads(item[1]))) # type: ignore

        return new_divs_and_dict


    def _offers_builder(self,):
        offers_dict: list[dict[str, str]] = [self._retrieve_json_content_as_dict(json_file) for json_file in self._json_files_explorer()]
        return offers_dict
    def detected_suppliers(self):
        """
        Retrieves the detected suppliers from the divs and offers data.
        Returns a list of unique suppliers with their relevant information.
        """

        suppliers: Sequence[SupplierDict] = list()
        console = rich.console.Console()
        with Progress(SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"), *Progress.get_default_columns(),console=console,transient=True) as progress:
            all_pages_task = progress.add_task(description="[blank]Parsing suppliers ...[/blank]")
            # single_page_task = progress.add_task(description="Retrieving suppliers ...", total=1)
            for divs, offers in self._divs_and_dict():
                    if offers['props']['offerResultData']['totalCount'] == 0:
                        continue 
                    for offer in offers['props']['offerResultData']['offers']: 
                        suppliers.append(
                                {
                                "name": offer['companyName'].lower(),   
                                "verified_type": suppliers_status(tags=divs, offer=offer),
                                "sopi_level": offer['displayStarLevel'],
                                "country_name": country_name(country_min=offer['countryCode']), 
                                "gold_supplier_year": offer['goldSupplierYears'].split(" ")[0], 
                                "supplier_service_score": float(offer["supplierService"])  
                            }
    
                        )
                    progress.update(all_pages_task, advance=100/len(self._divs_and_dict()))
        # removing supliers present twice in supliers dict
        unique_suppliers_tuple = list(
            OrderedDict((str(d), d) for d in suppliers).items()
        )
        # print(unique_suppliers_tuple)
        unique_suppliers = [
            supplier_tuple[1] for supplier_tuple in unique_suppliers_tuple
        ]
        return unique_suppliers

    def detected_products(self):
        """
        Returns a list of unique products with their relevant information.
        """
        products: Sequence[ProductDict] = list()
        console = rich.console.Console()
        with Progress(SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),*Progress.get_default_columns(),console=console,transient=True) as progress:
            all_pages_task = progress.add_task(description="[blank]Parsing products ...[/blank]") 
            for divs, offers in self._divs_and_dict():
                if offers['props']['offerResultData']['totalCount'] == 0:
                    continue 
                # print(f"total count : {offers['props']['offerResultData']['totalCount']}")
                for offer in offers['props']['offerResultData']['offers']: 
                    products.append(
                            {
                                "name": offer['enPureTitle'].lower(),
                                "max_price": get_product_price(all_price_text=offer['price'], which="max"),
                                "min_price": get_product_price(all_price_text=offer['price'], which="min"),
                                "guaranteed_by_alibaba": is_alibaba_guaranteed(str_status=offer['halfTrust']),
                                "certifications": get_product_certification(offer=offer),
                                "minimum_to_order": int(offer['halfTrustMoq'].lower()),
                                "ordered_or_sold": ordered_or_sold(offer=offer),
                                "supplied_by": offer['companyName'].lower(),
                                "product_score": float(offer["productScore"]),
                                "review_count": float(offer["reviewCount"]),
                                "review_score": float(offer["reviewScore"]),
                                "shipping_time_score":float(offer["shippingTime"]),
                                "is_full_promotion": is_full_promotion(str_status=offer["isFullPromotion"]),
                                "customizable" : is_customizable(str_status=offer["customizable"]),
                                "instant_order": is_instant_order(str_status=offer["halfTrustInstantOrder"]),
                                "trade_product": is_trade_product(str_status=offer['tradeProduct'])
                                
                            }
                        )
                progress.update(all_pages_task, advance=100/len(self._divs_and_dict()))
        unique_products_tuple = list(
            OrderedDict((str(d), d) for d in products).items()) 
        # removing supliers present twice in supliers dict
        unique_products = [product_tuple[1] for product_tuple in unique_products_tuple]
        return unique_products


if __name__ == "__main__":
    pp = PageParser(targeted_folder="Alibaba_Scapper_new")
    dt = pp.detected_suppliers()
    for d in dt:
        pprint(d)
    # help(pp)



import json
import pathlib
from pprint import pprint
import unicodedata
from typing import Literal

from click import Path
from loguru import logger
from selectolax.parser import Node


def _remove_dot_from_price(price_as_string: str):
    price = price_as_string.split(".")
    price_dot_removed = ".".join(price[:1])
    return float(price_dot_removed.strip("$US"))


@logger.catch(TypeError)
def get_product_price(all_price_text: str, which: Literal["max", "min"]):
    elt = all_price_text
    if "-" in elt and which == "max":
        max_price = elt.split("-")[1].split("&")[0]
        max_price = unicodedata.normalize("NFKC", max_price)
        max_price = max_price.replace("\u00a0", "").replace(",", ".").replace(" ", "")
        if max_price.count(".") > 1:
            max_price = _remove_dot_from_price(price_as_string=max_price)
            return max_price
        return float(max_price.strip("$US"))

    elif "-" in elt and which == "min":
        min_price = elt.split("-")[0].split("&")[0]
        min_price = unicodedata.normalize("NFKC", min_price)
        min_price = min_price.replace("\u00a0", "").replace(",", ".").replace(" ", "")
        if min_price.count(".") > 1:
            min_price = _remove_dot_from_price(price_as_string=min_price)
            return min_price
        return float(min_price.strip("$US"))
    else:
        strange_price = elt.strip(r"\xa0$US")
        price = unicodedata.normalize("NFKC", strange_price)
        price = price.replace("\u00a0", "").replace(",", ".").replace(" ", "")
        if price.count(".") > 1:
            price = _remove_dot_from_price(price_as_string=price)
            return price
        else:
            return float(price.strip("$US"))


@logger.catch(TypeError)
def is_alibaba_guaranteed(str_status: str):
    if str_status == "false":
        return False
    else:
        return True


@logger.catch(TypeError)
def get_product_certification(offer:dict):
    certifications_name = []
    for info in offer['productCertificates']:
        certifications_name.append(info["name"])
    
    return ",".join(certifications_name)

def is_full_promotion(str_status:str):
    if str_status == "false":
        return False
    else:
        return True
    
def is_customizable(str_status:str):
    if str_status == "false":
        return False
    else:
        return True
    
def is_instant_order(str_status:str):
    if str_status == "false":
        return False
    else:
        return True

def is_trade_product(str_status:str):
    if str_status == "false":
        return False
    else:
        return True

@logger.catch(TypeError)
def suppliers_status(tags: list[Node],offer:dict):
    for tag in tags :
        # print(tag.attributes)
        company_name = tag.css_first("a[data-spm='d_companyName']").text().strip().lower()
        # print(company_name+ "1")
        if company_name == offer['companyName'].lower().strip():
            # print(offer['companyName'].lower().strip()+"2")
            node_a = tag.css_first(".auth-info-group-normal.J_no_jump")
            if node_a is None:
                node_a = tag.css_first(".search-card-e-popper__trigger").child
                # print(node_a.attributes)
                node_a = node_a.css_first(".verified-supplier-icon__wrapper")
                if node_a is None:
                    return "unverified"
                mode = node_a.attributes.get("data-aplus-auto-card-mod")
                return mode.split("=")[2]
            node_a = node_a.css_first("a.verified-supplier-icon__wrapper")
            if node_a is not None:
                # print(node_a.attributes)
                mode = node_a.attributes.get("data-aplus-auto-card-mod")
                # print(mode)
                mode = mode.split("=")[2]
                return mode
            return "unverified" 
    return "unverified"
           


@logger.catch(TypeError)
def sopi_level(tag: Node):
    elt = tag.css_first(".search-cards-e-star")
    return len(elt.css(".search-card-e-iconfont"))


@logger.catch(TypeError)
def years_as_supplier_gold(tag: Node):
    elt = tag.css_first(".search-card-e-supplier__year")
    year = elt.attrs.get("data-aplus-auto-card-mod").split("@@")[1][0]
    return year


def _get_minimum_to_order_tag(tags: list[Node]):
    for tag in tags:
        element_attr = tag.attrs.get("data-aplus-auto-card-mod")
        # if 'price_negotiated' or "easy_return" not in element_attr:
        #     print(tag.text())
        #     return tag
        if "minimale" in element_attr:
            return tag
        # print(element_attr)


@logger.catch(TypeError)
def minimum_to_order(tag: Node):
    elements = tag.css(".search-card-m-sale-features__item.tow-line")
    if len(elements) != 0:
        element = _get_minimum_to_order_tag(tags=elements)
        if element is None:
            return 0
        number_str = element.text()
        number = number_str.split(":")[1].split()[0].strip()
        return int(number)
    else:
        return 0


@logger.catch(TypeError)
def ordered_or_sold(offer:dict):
    power_common_info = offer.get("marketingPowerCommon")
    if power_common_info is not None:
        return int( power_common_info['count'])
    else:
        return 0


def _from_abr_to_full_name(country_abr: str):
    p = (pathlib.Path(__file__).parent / "pays_data.json").resolve()
    with open(f"{p}", encoding="utf-8") as f:
        countries = json.load(
            f,
        )
        pays_data = countries["continents_pays"]
        for pays in pays_data:
            # print("two letter code : ", pays["Two_Letter_Country_Code"])
            # print("country abr : ", country_abr)
            if pays["Two_Letter_Country_Code"].lower() == country_abr.lower():
                # print("matched country : ", country_abr)
                return pays["Country_Name"].lower()
            if country_abr.lower() == "uk":
                return "royaume-uni"
        return "unknow"


def country_name(country_min: str):
    return _from_abr_to_full_name(country_abr=country_min)



"""
This script is responsible for scraping data from the Alibaba website.
It uses the Playwright library to navigate through multiple pages
extract HTML content,and save it to disk. The script defines
functions to handle asynchronous tasks,interact with
browser contexts, and gather data from the scraped HTML files.
Additionally, it utilizes loguru for logging purposes and decouple for
managing environment variables.
"""

import asyncio
from pprint import pprint
import sys
import time
from tkinter import Place



from click import UsageError
import playwright.sync_api
import requests
import playwright
import selectolax
from playwright.sync_api import Error as PError
from playwright.async_api import Error as APError
from dotenv import dotenv_values, load_dotenv
import urllib3
from .html_to_disk import write_to_disk
from .info_message import run_scrapper_with_success
from loguru import logger
from playwright.async_api import Page as AsyncPage
from playwright.async_api import async_playwright, BrowserContext
from playwright.sync_api import sync_playwright
from rich import print as rprint
from rich.progress import Progress, TaskID
from rich.console import Console
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn, TimeElapsedColumn,SpinnerColumn
import os
from tenacity import AsyncRetrying, RetryError, retry, stop_after_attempt
from . import SBR_WS_CDP_LIST
# SECRETS_KEYS = dotenv_values(".env")
HTML_PAGE_RESULT = []



def _browser_parser(html_content: str | bytes, curr_url: str):
    """Parse the HTML content of the page to get the divs with class `.organic-list.viewtype-list`

    :param html_content: The HTML content of the page
    :type html_content: str
    :return: A list of divs with class `.organic-list.viewtype-list`
    :rtype: list
    """
    if html_content is None:
        logger.warning(
            " your page could not be loaded, an empty none value has been returned "
        )
        return None
    body_parser = selectolax.parser.HTMLParser(html_content)
    logger.info(
        f"looking for .organic-list class in the DOM page {curr_url.split('page=')[1]}... "
    )
    product_div = body_parser.css_first(".organic-list")
    if product_div is None:
        return None
    html_result = product_div.html
    if html_result is None:
        logger.warning(
            f"any HTML content from page {curr_url.split('page=')[1]} matched the selector '.organic-list', None value has been returned"
        )
        return None
    else:
        logger.info(
            f"HTML content from page {curr_url.split('page=')[1]} has been returned  "
        )
        return html_result

@retry(stop=stop_after_attempt(3))
async def goto_task(
    url: str,
    context_browser: BrowserContext,
    task: TaskID,
    progress: Progress,
    semaphore: asyncio.Semaphore,
    tg_instance:asyncio.TaskGroup,
    page:AsyncPage,
    page_results:int
) -> str | None:
    """Return the entire HTML content from each page only the divs with class `.organic-list.viewtype-list`

    :param url: URL of page range 1 to 42 include
    :type url: str
    :param context_browser: The browser context to use
    :type context_browser: BrowserContext
    :return: A string of HTML content of the current page, or None if the page has no expected div
    :rtype: Optional[str]
    """
    # page = await context_browser.new_page()
    try:
        async with page:
            logger.info(f"Loading page {url.split('page=')[1]} ... ")
            response = await page.goto(url, wait_until="domcontentloaded", timeout=0)
            if response is None:
                return None
            logger.info(f"get response text from web page {url.split('page=')[1]} ... ")
            html_body = await response.text()
            progress.start_task(task)
            progress.update(task, advance=100/page_results)
            global HTML_PAGE_RESULT
            HTML_PAGE_RESULT.append(html_body)
            logger.info(f"Closing page {url.split('page=')[1]} ... ")
    except Exception as e:
        logger.error(f"Error processing page {url.split('page=')[1]}: {e}")
        return None

def urls_pusher(words:str,stop_at:int):
    for i in range(1,stop_at+1):
        yield f"https://french.alibaba.com/trade/search?spm=a2700.galleryofferlist.0.0.538f1619xDyhzJ&fsb=y&IndexArea=product_en&keywords={words}&tab=all&&page={i}"

@logger.catch()
async def async_scrapper(*, save_in: str, key_words: str,page_results:int) -> None:
    """Create a list of tasks with `goto` coroutine and pass it to `asyncio.wait` coroutine
    then wait for all the results.
    The function takes no arguments and returns None.
    :return: None
    """
    async with async_playwright() as p:
        logger.info("Connecting to CDP and creating the browser... ")
        try:
            response = requests.get("http://geo.brdtest.com/mygeo.json")
            # print(response.raise_for_status())
            country_name = response.json()['country']
            api_key = SBR_WS_CDP_LIST.replace("country-us", f"country-{country_name.lower()}")
            browser = await p.chromium.connect_over_cdp(api_key)
        except urllib3.exceptions.NameResolutionError :
            raise UsageError("check your internet connection")
        
        except playwright._impl._errors.Error as e:  # type: ignore
            if "Account is suspended" in str(e):
                # print(str(e))
                rprint(
                    "[red] Bright data account has been suspended by system. contact me by email: [magenta]onealzero@gmail.com [/magenta] to fix this as soon as possible. You can use `--sync-api` flag after your last command to enable sync scrapping instead while i'm fixing bug.  [/red]"
                )
                return
            elif "exists" in str(e):
                # print(str(e))
                rprint("[white] Seems like playwright is not installed. lets aba install it for you... [/white]")
                os.system("playwright install")
            else:
                print(str(e))

        context_browser = await browser.new_context()
        with Progress(SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),*Progress.get_default_columns(),transient=True) as progress:
            task = progress.add_task("[green blink] async Scraping...", start=False)
            s_one = asyncio.Semaphore(value=10)
            logger.info("Loading pages results ... ")
            async with asyncio.Lock() :
                async with asyncio.TaskGroup() as tg:
                    for url in urls_pusher(words=key_words,stop_at=page_results):
                        async with asyncio.Lock():
                            page = await context_browser.new_page()
                        tg.create_task(goto_task(url=url, semaphore=s_one, progress=progress, task=task,tg_instance=tg,context_browser=context_browser,page=page,page_results=page_results))   
                        
        write_to_disk(save_in,HTML_PAGE_RESULT)
        run_scrapper_with_success(folder_name=save_in)


def sync_scrapper(*, save_in: str, key_words: str,page_results:int) -> None:

    # pages_urls = _looking_for_urls(keywords=key_words)
    with Progress(SpinnerColumn(finished_text="[bold green]finished ✓[/bold green]"),*Progress.get_default_columns(),transient=True) as progress:
        task = progress.add_task("[green blink] Sync Scraping...", start=False,)
        playwright = sync_playwright().start()
        try:    
            browser = playwright.chromium.launch(headless=True)
        except playwright._impl._errors.Error:  # type: ignore
            rprint("[white] Seems like playwright is not installed. lets aba install it for you... [/white]")
            os.system("playwright install")
        context = browser.new_context()
        for url in urls_pusher(words=key_words,stop_at=page_results):
            page = context.new_page()
            logger.info(f"Loading page {url.split('page=')[1]} ... ")
            try:
                response = page.goto(url, wait_until="domcontentloaded", timeout=0)
                if response is None:
                    return None
                logger.info(
                    f"Returns the text representation of response body from page {url.split('page=')[1]} ... "
                )
            except PError as e: 
                if "ERR_INTERNET_DISCONNECTED" in e.message:
                    raise UsageError("Check your internet connection ... ")
            html_content = response.text()
            progress.update(task, advance=100/page_results)
            global HTML_PAGE_RESULT
            HTML_PAGE_RESULT.append(html_content)
            logger.info(f"Closing the page {url.split('page=')[1]} ... ")
            page.close()
    write_to_disk(save_in, HTML_PAGE_RESULT)
    run_scrapper_with_success(folder_name=save_in)


if __name__ == "__main__":
    start_time = time.perf_counter()
    # asyncio.run(async_scrapper(save_in='pc_lenovo', key_words='pc lenovo'),debug=False)
    sync_scrapper(save_in="pc_lenovo", key_words="pc lenovo", page_results=30)
    end_time = time.perf_counter()
    print(f" all those tasks tooks: {end_time-start_time:.2f}")
    ...



from typing import TypedDict

class ProductDict(TypedDict):
    name: str 
    guaranteed_by_alibaba: bool
    certifications: str
    minimum_to_order: int
    ordered_or_sold: int
    min_price: float
    max_price: float
    product_score: float
    review_count : float
    review_score : float
    shipping_time_score : float
    is_full_promotion: bool
    customizable: bool
    instant_order: bool
    trade_product:bool
    supplied_by : str

class SupplierDict(TypedDict):
    name: str 
    verified_type: str
    sopi_level: int
    country_name: str
    gold_supplier_year: int
    supplier_service_score: float